import {getScrollLeft, getScrollTop} from 'vhtml-ui/src/utils/dom';
import {dashcase} from 'vhtml-ui/src/utils/lang';

const CONTEXTMENU_HOOK_PROPNAME = 'VHTML_CONTEXTMENU_HOOK_PROPNAME';

/**
 * 查找contextmenu
 * @param {VueContext} context 上下文
 * @param {string} refName 引用名字
 * @param {string} componentName 组件名字
 * @return {Component} contextmenu
 */
function getContextmenu(context, refName, componentName) {
    let contextmenu = context.$refs[refName];
    if (Array.isArray(contextmenu)) {
        // 如果有多个，则用第一个
        // 有多个相同`ref`或者在`v-for`上使用`ref`会导致有多个
        contextmenu = contextmenu[0];
    }

    while (contextmenu && contextmenu.$options.name !== 'VContextmenu') {
        // 其他基于`VContextmenu`扩展的组件，可以再找一层，找到真正的`v-contextmenu`
        contextmenu = contextmenu.$refs.contextmenu;
    }

    if (!contextmenu) {
        throw new Error(`[VHTML-UI][${dashcase(componentName)}] ${componentName} "${refName}" not found.`);
    }

    return contextmenu;
}

function unbindDirective(el) {
    /**
     * @type {Object} options
     * @type {Object} options.onContextmenu onContextmenu callback
     * @type {Object} options.unbindCallback unbindCallback
     */
    const options = el[CONTEXTMENU_HOOK_PROPNAME];
    if (!options) {
        return;
    }
    el.removeEventListener('contextmenu', options.onContextmenu);

    // hide popup
    options.contextmenu.hide();
    options.contextmenu = null;
    el[CONTEXTMENU_HOOK_PROPNAME] = null;
}

function createBindDirective(componentName) {
    return function (el, binding, vnode) {
        const context = vnode.context;
        // dynamic value > argument
        const arg = binding.arg;
        const value = binding.value;
        const refName = arg || value;

        if (!refName || typeof refName !== 'string') {
            return;
        }

        let thisContextmenu = getContextmenu(context, refName, componentName);
        function onContextmenu(e) {
            e.stopPropagation();
            thisContextmenu.$emit('beforeShow', e);
            if (e.defaultPrevented) {
                return;
            }
            e.preventDefault();

            const menu = thisContextmenu.$refs.contextmenu;
            const pageX = e.pageX;
            const pageY = e.pageY;
            let left = pageX - getScrollLeft();
            let top = pageY - getScrollTop();

            menu.style.left = left + 'px';
            menu.style.top = top + 'px';
            thisContextmenu.show();
        }

        const defineValue = {
            contextmenu: thisContextmenu,
            onContextmenu
        };
        const oldOptions = el[CONTEXTMENU_HOOK_PROPNAME];
        if (oldOptions) {
            /*
            if (oldOptions.popper.currentVisible) {
                // if the old popper visible
                // then show the new popper
                popper.setTrigger(triggerElement);
                popper.show();
            }
            */
            unbindDirective(el);
            el[CONTEXTMENU_HOOK_PROPNAME] = defineValue;
        }
        else {
            Object.defineProperty(el, CONTEXTMENU_HOOK_PROPNAME, {
                value: defineValue,
                enumerable: false,
                configurable: true,
                writable: true
            });
        }

        // unbind on popper destroy
        thisContextmenu.$on('beforeDestroy', () => {
            unbindDirective(el);
        });

        el.addEventListener('contextmenu', onContextmenu, false);
    };
}

export default function createPopperDirective(componentName) {
    const bindDirective = createBindDirective(componentName);
    return {
        bind: bindDirective,
        unbind: unbindDirective
    };
}
