import Dialog from './src/dialog';

const createDialogContianer = function () {
    const div = document.createElement('div');
    document.body.appendChild(div);
    return div;
};
const removeDialog = function (dialog) {
    dialog.$destroy();
    if (dialog.$el && dialog.$el.parentNode) {
        dialog.$el.parentNode.removeChild(dialog.$el);
    }
};

const confirmFactory = function (Vue) {
    /**
     * @param {Object} options options
     * @param {string} options.title title
     * @param {string} options.content content message
     * @param {function} options.render render function to render more complex view
     * @param {string} options.status status,
     *                  @eg: 'primary', 'success', 'warning'
     *                  default is 'primary'
     * @param {string} options.confirmText text of confirm button
     * @param {string} options.cancelText text of cancel button
     * @param {function(Event)} options.confirm callback of confirm button click
     * @param {function(Event)} options.cancel callback of cancel button click
     * @param {function(Event)} options.close callback of close icon click
     * @return {Vue} dialog component instance
     */
    return function $confirm(options) {
        const container = createDialogContianer();
        const dialog = new Vue({
            el: container,
            data() {
                return {
                    confirmLoading: false
                };
            },
            render(createElement, context) {
                return createElement(
                    'v-dialog',
                    {
                        ref: 'dialog',
                        props: {
                            open: true,
                            kind: 'confirm',
                            status: options.status || 'primary',
                            title: options.title,
                            confirmLoading: this.confirmLoading
                        },
                        on: {
                            confirm(e) {
                                if (typeof options.confirm === 'function') {
                                    options.confirm(e);
                                }

                                if (!e.defaultPrevented) {
                                    removeDialog(dialog);
                                }
                            },
                            cancel(e) {
                                if (typeof options.cancel === 'function') {
                                    options.cancel(e);
                                }

                                if (!e.defaultPrevented) {
                                    removeDialog(dialog);

                                }
                            },
                            close(e) {
                                if (typeof options.close === 'function') {
                                    options.close(e);
                                }

                                if (!e.defaultPrevented) {
                                    removeDialog(dialog);
                                }
                            }
                        }
                    },
                    [
                        options.render
                            ? options.render(createElement)
                            : options.content
                                ? createElement('div', {
                                    domProps: {
                                        innerHTML: options.content
                                    }
                                })
                                : '',
                        options.confirmText ? createElement('span', {slot: 'confirm-text'}, options.confirmText) : '',
                        options.cancelText ? createElement('span', {slot: 'cancel-text'}, options.cancelText) : ''
                    ]
                );
            }
        });
        dialog.$remove = function () {
            removeDialog(this);
        };
        dialog.showButtonLoading = function () {
            this.confirmLoading = true;
        };
        dialog.hideButtonLoading = function () {
            this.confirmLoading = false;
        };

        return dialog;
    };
};

const alertFactory = function (Vue) {
    /**
     * @param {Object} options options
     * @param {string} options.title title
     * @param {string} options.content content message
     * @param {function} options.render render function to render more complex view
     * @param {string} options.status status,
     *                  @eg: 'primary', 'success', 'warning'
     *                  default is 'primary'
     * @param {string} options.button-text text of confirm button
     * @param {Funtion} options.close triggered on dialog closed
     */
    return function $alert(options) {
        const container = createDialogContianer();
        const dialog = new Vue({
            el: container,
            render(createElement, context) {
                return createElement(
                    'v-dialog',
                    {
                        ref: 'dialog',
                        props: {
                            open: true,
                            kind: 'alert',
                            status: options.status || 'primary',
                            title: options.title
                        },
                        on: {
                            close(e) {
                                if (typeof options.close === 'function') {
                                    options.close(e);
                                }

                                if (!e.defaultPrevented) {
                                    removeDialog(dialog);
                                }
                            }
                        }
                    },
                    [
                        options.render
                            ? options.render(createElement)
                            : options.content
                                ? createElement('div', {
                                    domProps: {
                                        innerHTML: options.content
                                    }
                                })
                                : '',
                        options.buttonText ? createElement('span', {slot: 'button-text'}, options.buttonText) : ''
                    ]
                );
            }
        });
    };
};

/* istanbul ignore next */
Dialog.install = function (Vue) {
    Vue.component(Dialog.name, Dialog);
    Vue.prototype.$confirm = confirmFactory(Vue);
    Vue.prototype.$alert = alertFactory(Vue);
};

export default Dialog;
