import {select, deselect} from '../util/range';
import {getRangeInsideElement} from '../util/range';
import {firstTextNode, deepest} from '../util/find';
import {getText} from '../util/dom';
import closest from '../util/closest';
import Observer from './observer';
import FallbackCompositionState from './fallback-composition-state';
import {USE_FALLBACK_COMPOSITION_DATA, IS_CHROME} from './constants/env';

import * as beforeinput from './events/beforeinput';
import * as compositionend from './events/compositionend';
import * as compositionstart from './events/compositionstart';
import * as copy from './events/copy';
import * as cut from './events/cut';
import * as dragover from './events/dragover';
import * as drop from './events/drop';
import * as keydown from './events/keydown';
import * as keypress from './events/keypress';
import * as mousedown from './events/mousedown';
import * as mouseup from './events/mouseup';
import * as paste from './events/paste';
import * as selectionchange from './events/selectionchange';
import * as textinput from './events/text-input';
const eventHandlers = [
    beforeinput, compositionend, compositionstart,
    copy, cut, paste, dragover, drop,
    keydown, keypress,
    mousedown, mouseup,
    selectionchange, textinput
];

export default class DomState {
    constructor(el, state) {
        this.el = el;
        this.state = state;
        // 是否输入的是空格键
        this.hasSpaceKeypress = false;
        // 是否正处于输入法合成过程中
        this.inComposing = false;
        this.bindEvents();
        this.observer = new Observer(this);
        this.observer.bind();
        if (USE_FALLBACK_COMPOSITION_DATA) {
            this.fallbackCompositionState = new FallbackCompositionState();
        }
    }

    destroy() {
        if (this.restoreSelectionTimer) {
            clearTimeout(this.restoreSelectionTimer);
        }
        this.bindEvents();
        this.observer.unbind();
    }

    bindEvents() {
        eventHandlers.forEach((eventConfig) => {
            if (eventConfig.NOT_BIND) {
                return;
            }

            const observer = this;
            eventConfig.trueCallback = function (event) {
                eventConfig.callback.call(this, event, observer);
            };
            const element = eventConfig.ON_DOCUMENT ? this.el.ownerDocument : this.el;
            element.addEventListener(eventConfig.name, eventConfig.trueCallback, false);
        });
    }

    unbindEvents() {
        eventHandlers.forEach((eventConfig) => {
            if (!eventConfig.trueCallback) {
                return;
            }
            this.el.removeEventListener(eventConfig.name, eventConfig.trueCallback);
        });
    }

    onInsertText(text) {
        if (!text || typeof text !== 'string') {
            return;
        }

        // 记录插入之前，值是否是空的
        // 如果是空的，就需要延迟restoreSelection的执行时间节点（Chrome下独有的bug）
        this.delayRestoreSelection = this.state.isEmpty();
        this.state.save('insertText');
        this.state.insertText(text);
    }

    onDelSelection() {
        this.state.save('delSelection');
        this.state.deleteSelection();
    }

    onBackspace() {
        this.state.save('backspace');
        this.state.backspace();
    }

    onDel() {
        this.state.save('del');
        this.state.del();
    }

    onEnter() {
        // this.state.save('enter');
        // this.state.insertNewLine();
    }

    onCut() {
        this.state.save('cut');
        const cuttedText = this.state.getSelectedText();
        this.state.deleteSelection();
        return cuttedText;
    }

    onPaste(text) {
        this.state.save('paste');
        this.state.deleteSelection();
        this.state.insertText(text);
    }

    onBackward(e) {
        this.state.backward();
    }

    onForward(e) {
        this.state.forward();
    }

    onUndo() {
        this.state.undo();
    }

    onRedo() {
        this.state.redo();
    }

    nearestKeyElement(container, offset) {
        const node = closest(container, '[data-key]', this.el);
        if (node) {
            return {node, offset};
        }

        // 找前一个和后一个节点树上面有没有data-key
        const isDataKey = function (node) {
            return !!node.getAttribute('data-key');
        };
        let kind;
        let beforeChildNode = container.childNodes[offset - 1];
        beforeChildNode = beforeChildNode
            ? closest(beforeChildNode, '[data-key]', this.el)
                || deepest(beforeChildNode, 'lastElementChild', isDataKey)
            : null;
        if (beforeChildNode) {
            kind = beforeChildNode.getAttribute('data-kind');
            return beforeChildNode
                ? {node: beforeChildNode, offset: kind === 'text' ? getText(beforeChildNode).length : 0}
                : null;
        }

        let afterChildNode = container.childNodes[offset];
        afterChildNode = afterChildNode
            ? closest(afterChildNode, '[data-key]', this.el)
                || deepest(afterChildNode, 'firstElementChild', isDataKey)
            : null;
        if (afterChildNode) {
            kind = afterChildNode.getAttribute('data-kind');
            return afterChildNode
                ? {node: afterChildNode, offset: 0}
                : null;
        }

        // 针对ctrl-A全选的情况处理，此时startContainer是根节点，没有[data-key]
        // 需要针对开头和结尾做特殊处理
        /*
        if (offset === 0) {
            const node = deepest(this.el, 'firstElementChild', isDataKey);
            return node ? {node, offset} : null;
        }
        else if (offset === container.childNodes.length) {
            const node = deepest(this.el, 'lastElementChild', isDataKey);
            const kind = node.getAttribute('data-kind');
            return node ? {node, offset: kind === 'text' ? getText(node).length : 0} : null;
        }
        */
    }

    applyRange(range) {
        const startAnchor = this.nearestKeyElement(range.startContainer, range.startOffset);
        if (!startAnchor) {
            this.state.deselect();
            return;
        }

        const endAnchor = this.nearestKeyElement(range.endContainer, range.endOffset);
        if (!endAnchor) {
            this.state.deselect();
            return;
        }

        const startKey = startAnchor.node.getAttribute('data-key');
        const start = this.state.findNode(startKey);
        if (!start) {
            this.state.deselect();
            return;
        }

        let endKey;
        let endOffset;
        // 当文字内容为空时强制设置为0，因为在展示时会把空文本替换成特殊空格，所以range.startOffset会是1
        const startOffset = start.node.kind === 'text' && start.node.text ? startAnchor.offset : 0;
        if (range.isCollapsed) {
            endKey = startKey;
            endOffset = startOffset;
        }
        else {
            endKey = endAnchor.node.getAttribute('data-key');
            const end = this.state.findNode(endKey);
            if (!end) {
                // NOTE 这里将来可能会有问题，没有结束节点则认为他是开始节点
                endKey = startKey;
                endOffset = startOffset;
            }
            else {
                endOffset = end.node.kind === 'text' && end.node.text ? endAnchor.offset : 0;
            }
        }

        this.state.updateSelection(startKey, startOffset, endKey, endOffset);
    }

    onSelectionChange() {
        if (this.inComposing || !this.state.selectionApplyed || this.observer.dirty) {
            return;
        }

        const range = getRangeInsideElement(this.el);
        if (!range) {
            this.state.deselect();
            return;
        }

        this.applyRange(range);
    }

    forceRestoreSelection() {
        const selection = this.state.selection;
        if (!this.state.isSelected()) {
            deselect(this.el);
            return;
        }

        const startDomNode = this.el.querySelector(`[data-key="${selection.startKey}"]`);
        const endDomNode = this.el.querySelector(`[data-key="${selection.endKey}"]`);
        if (!startDomNode || !endDomNode) {
            return;
        }

        const startKind = startDomNode.getAttribute('data-kind');

        const endKind = endDomNode.getAttribute('data-kind');
        const startContainer = startKind === 'text'
            ? startDomNode.firstChild
            : firstTextNode(startDomNode);
        const startOffset = selection.startOffset;
        const endContainer = endKind === 'text'
            ? endDomNode.firstChild
            : firstTextNode(endDomNode);
        const endOffset = selection.endOffset;
        if (!startContainer || !endContainer) {
            return;
        }

        select(startContainer, startOffset, endContainer, endOffset);
    }

    restoreSelection() {
        if (this.state.selectionApplyed || this.isRestoring) {
            return;
        }

        this.isRestoring = true;
        const restore = () => {
            this.isRestoring = false;
            this.state.selectionApplyed = true;
            this.delayRestoreSelection = false;
            this.restoreSelectionTimer = null;
            this.forceRestoreSelection();
        };

        if (IS_CHROME && this.delayRestoreSelection) {
            // chrome下有奇怪的bug，中文输入法在空内容的时候输入中文，
            // 会导致光标restore之后，立即被定位到其他地方（表现为光标消失）
            // 故需要延迟光标restore时间节点
            this.restoreSelectionTimer = setTimeout(restore, 0);
        }
        else {
            restore();
        }
    }
}
