import omit from 'lodash/omit';
import cloneDeep from 'lodash/cloneDeep';
import isEqual from 'lodash/isEqual';
import history from './history';

let keyCount = 0;

function getDefaultValue(customValue, obj, key, defaultValue) {
    return customValue != null ? customValue : (obj && obj[key] || defaultValue);
}

export default class EditorState {
    constructor() {
        this.reset();
    }

    reset() {
        /**
         *  [
         *      {kind: 'text', key: '1', text: ''},
         *      {kind: 'block', key: '2', ......}
         *  ]
         *
         */
        this.lines = [[this.createText()]];
        /**
         *  {
         *      startKey: '1',
         *      startOffset: 10,
         *      endKey: '2',
         *      endOffset: 0
         *  }
         *  默认为不选中状态
         */
        this.selection = {
            selected: false,
            startKey: '',
            startOffset: 0,
            endKey: '',
            endOffset: 0
        };
        this.selectionApplyed = true;
        this.selectedNoneTextNode = [];
        // blur之前记录的选中状态snapshot
        this.savedSnapshotSelection = null;
        // undo 和 redo
        this.undoStack = [];
        this.redoStack = [];
    }

    normalize(row) {
        if (typeof row === 'number') {
            const line = this.lines[row];
            if (!line) {
                return;
            }

            let col = 1;
            while (col < line.length) {
                const beforeNode = line[col - 1];
                const afterNode = line[col];
                if (beforeNode.kind === 'text' && afterNode.kind === 'text') {
                    // 更新selection
                    if (this.selection.startKey === afterNode.key) {
                        this.select(beforeNode.key, this.selection.startOffset + beforeNode.text.length);
                    }
                    if (this.selection.endKey === afterNode.key) {
                        this.select(null, null, beforeNode.key, this.selection.endOffset + beforeNode.text.length);
                    }
                    // 强制刷新选中
                    this.selectionApplyed = false;
                    // 合并前后节点
                    beforeNode.text += afterNode.text;
                    line.splice(col, 1);
                }
                else {
                    col++;
                }
            }
            return;
        }

        for (let row = 0, rowL = this.lines.length; row < rowL; row++) {
            this.normalize(row);
        }
    }

    replaceNodeKey(node, selection) {
        const oldKey = node.key;
        const newKey = this.generateKey();
        node.key = newKey;

        if (!selection.selected) {
            return;
        }

        if (selection.startKey === oldKey) {
            selection.startKey = newKey;
        }
        if (selection.endKey === oldKey) {
            selection.endKey = newKey;
        }
    }

    isSelected() {
        return !!this.selection.selected;
    }

    isCollapsed() {
        if (!this.isSelected()) {
            return false;
        }
        const {startKey, startOffset, endKey, endOffset} = this.selection;
        return startKey === endKey && startOffset === endOffset;
    }

    selectAtLast() {
        const node = this.lastNode();
        this.setCaret(node.key, node.kind === 'text' ? node.text.length : 0);
    }

    deselect() {
        if (!this.isSelected()) {
            // 相同
            return;
        }
        this.selection.selected = false;
        this.selection.startKey = this.selection.endKey = '';
        this.selection.startOffset = this.selection.endOffset = 0;
        this.selectionApplyed = false;
        this.updateSelectedNoneTextNode();
    }

    // 清除所有选中状态
    clearSelection() {
        // 取消选中
        this.deselect();
        // 同时移除blur时保存的光标位置记录
        this.savedSnapshotSelection = null;
    }

    // 已知了dom的选中状态，同步到editorState中
    updateSelection(startKey, startOffset, endKey, endOffset) {
        const oldSelection = this.selection;
        startKey = getDefaultValue(startKey, oldSelection, 'startKey', '');
        startOffset = getDefaultValue(startOffset, oldSelection, 'startOffset', 0);
        endKey = getDefaultValue(endKey, oldSelection, 'endKey', '');
        endOffset = getDefaultValue(endOffset, oldSelection, 'endOffset', 0);

        if (!oldSelection
            || oldSelection.startKey !== startKey
            || oldSelection.startOffset !== startOffset
            || oldSelection.endKey !== endKey
            || oldSelection.endOffset !== endOffset) {
            // 不相同，避免是代码操作过程中的移动导致了selectionchange事件
            this.stopIgnoreNextSave();
        }

        this.selectionApplyed = true;
        this.selection.selected = true;
        this.selection.startKey = startKey;
        this.selection.startOffset = startOffset;
        this.selection.endKey = endKey;
        this.selection.endOffset = endOffset;
        this.updateSelectedNoneTextNode();
    }

    // 更新editorState的选中，并且之后会在组件update之后dom选中状态
    select(startKey, startOffset, endKey, endOffset) {
        const oldSelection = this.selection;
        startKey = getDefaultValue(startKey, oldSelection, 'startKey', '');
        startOffset = getDefaultValue(startOffset, oldSelection, 'startOffset', 0);
        endKey = getDefaultValue(endKey, oldSelection, 'endKey', '');
        endOffset = getDefaultValue(endOffset, oldSelection, 'endOffset', 0);

        if (oldSelection
            && oldSelection.startKey === startKey
            && oldSelection.startOffset === startOffset
            && oldSelection.endKey === endKey
            && oldSelection.endOffset === endOffset) {
            // 相同
            return;
        }

        this.selectionApplyed = false;
        this.selection.selected = true;
        this.selection.startKey = startKey;
        this.selection.startOffset = startOffset;
        this.selection.endKey = endKey;
        this.selection.endOffset = endOffset;
        this.updateSelectedNoneTextNode();
    }

    updateSelectedNoneTextNode() {
        this.selectedNoneTextNode.splice(0, this.selectedNoneTextNode.length);
        if (!this.isSelected()) {
            return;
        }

        this.eachNodeInSelection((node) => {
            if (node.kind === 'text') {
                return;
            }
            this.selectedNoneTextNode.push(node.key);
        });
    }

    eachNodeInSelection(callback) {
        if (!this.isSelected()) {
            return;
        }

        const start = this.findStartNode();
        const end = this.findEndNode();
        for (let row = start.row; row <= end.row; row++) {
            const line = this.lines[row];
            if (!line) {
                break;
            }
            const isStartLine = row === start.row;
            const isEndLine = row === end.row;
            let col = isStartLine ? start.col : 0;
            let currentColEnd = isEndLine ? end.col : line.length - 1;
            for (; col <= currentColEnd; col++) {
                const node = line[col];
                if (!node) {
                    break;
                }
                callback(
                    node, row, col,
                    start.row === row && start.col === col,     // is start node
                    end.row === row && end.col === col          // is end node
                );
            }
        }
    }

    setCaret(key, offset) {
        this.select(key, offset, key, offset);
    }

    setCaretAtLast(node) {
        if (!node) {
            return;
        }

        const key = node.key;
        const offset = node.kind === 'text' ? node.text.length : 0;
        this.select(key, offset, key, offset);
    }

    setCaretAtFirst(node) {
        if (!node) {
            return;
        }

        const key = node.key;
        const offset = 0;
        this.select(key, offset, key, offset);
    }

    backward() {
        if (!this.isSelected()) {
            return;
        }

        const {startOffset, startKey} = this.selection;
        if (!this.isCollapsed()) {
            // 不是光标状态，则变成光标状态
            this.setCaret(startKey, startOffset);
            return;
        }

        const start = this.findNode(startKey);
        if (start.node.kind === 'text' && startOffset > 0) {
            // 文字节点内部移动
            this.setCaret(startKey, startOffset - 1);
            return;
        }

        // 移动到前一个节点最后
        const prev = this.prevNode(start);
        if (!prev) {
            // 前面没有节点不能再移动
            return;
        }
        this.setCaretAtLast(prev.node);
    }

    forward() {
        if (!this.isSelected()) {
            return;
        }

        const {endOffset, endKey} = this.selection;
        if (!this.isCollapsed()) {
            // 不是光标状态，则变成光标状态
            this.setCaret(endKey, endOffset);
            return;
        }

        const end = this.findNode(endKey);
        if (end.node.kind === 'text' && endOffset < end.node.text.length) {
            // 文字节点内部移动
            this.setCaret(endKey, endOffset + 1);
            return;
        }

        // 移动到后一个节点最前
        const next = this.nextNode(end);
        if (!next) {
            // 后面没有节点不能再移动
            return;
        }
        this.setCaretAtFirst(next.node);
    }

    collapse(toStart) {
        const currentSelection = this.selection;
        if (!currentSelection) {
            return;
        }

        if (currentSelection.startKey === currentSelection.endKey
            && currentSelection.startOffset === currentSelection.endOffset) {
            // 已经collapse了
            return;
        }

        if (toStart) {
            currentSelection.endKey = currentSelection.startKey;
            currentSelection.endOffset = currentSelection.startOffset;
        }
        else {
            currentSelection.startKey = currentSelection.endKey;
            currentSelection.startOffset = currentSelection.endOffset;
        }
        this.selectionApplyed = false;
        this.updateSelectedNoneTextNode();
    }

    getSelectedText() {
        if (!this.isSelected() || this.isCollapsed()) {
            // 没有选中
            return '';
        }

        const {startOffset, endOffset} = this.selection;
        let text = '';
        this.eachNodeInSelection((node, row, col, isStartNode, isEndNode) => {
            if (node.kind !== 'text') {
                return;
            }

            text += !isStartNode && !isEndNode
                ? node.text
                : node.text.slice(isStartNode ? startOffset : 0, isEndNode ? endOffset : node.text.length);
        });
        return text;
    }

    deleteSelection() {
        if (!this.isSelected() || this.isCollapsed()) {
            // 没有选中
            return;
        }

        const start = this.findNode(this.selection.startKey);
        const end = this.findNode(this.selection.endKey);
        if (!start || !end) {
            return;
        }

        const removeTextOrNode = (startOrEnd, textStart, textEnd) => {
            if (startOrEnd.node.kind === 'text') {
                startOrEnd.node.text = startOrEnd.node.text.slice(textStart, textEnd);
            }
            else {
                const newTextNode = this.createText();
                this.lines[startOrEnd.row].splice(startOrEnd.col, 1, newTextNode);
                if (startOrEnd.node.key === this.selection.startKey) {
                    // 如果开始节点被删掉了，则移动到前面
                    this.selection.startKey = newTextNode.key;
                    this.selection.startOffset = 0;
                }
            }
        };

        if (start.node === end.node) {
            // 同一个节点，非文本节点不可能存在这个情况
            const startNodeText = start.node.text;
            start.node.text = startNodeText.slice(0, this.selection.startOffset)
                + startNodeText.slice(this.selection.endOffset);
            this.collapse(true);
        }
        else if (end.row === start.row) {
            // 同一行不同节点
            // 删除结束节点
            removeTextOrNode(end, this.selection.endOffset);
            // 删除开始和结束节点之间的节点
            const startCol = start.col + 1;
            const endCol = end.col;
            if (startCol < endCol) {
                this.lines[start.row].splice(startCol, endCol - startCol);
            }
            // 删除开始节点
            removeTextOrNode(start, 0, this.selection.startOffset);
            // normalize
            this.collapse(true);
            this.normalize(start.row);
        }
        else {
            // 不同行
            // 删除结束节点
            removeTextOrNode(end, this.selection.endOffset);
            // 删除开始和结束行的多余节点
            const startCol = start.col + 1;
            const startLine = this.lines[start.row];
            const endLine = this.lines[end.row];
            startLine.splice(startCol, startLine.length - startCol);
            endLine.splice(0, end.col);
            // 删除开始和结束行之间的多余节点
            const startRow = start.row + 1;
            const endRow = end.row;
            if (startRow < endRow) {
                this.lines.splice(startRow, endRow - startRow);
            }
            // 合并两行
            const targetLine = this.lines[start.row];
            const srcLine = this.lines[start.row + 1];
            targetLine.push.apply(targetLine, srcLine);
            this.lines.splice(start.row + 1, 1);
            // 删除开始节点
            removeTextOrNode(start, 0, this.selection.startOffset);
            // normalize
            this.collapse(true);
            this.normalize(start.row);
        }
    }

    findNode(key) {
        if (!key) {
            return null;
        }

        let row;
        let col;
        for (row = 0; row < this.lines.length; row++) {
            const line = this.lines[row];
            for (col = 0; col < line.length; col++) {
                if (line[col].key === key) {
                    return {
                        row,
                        col,
                        node: line[col]
                    };
                }
            }
        }
        return null;
    }

    findStartNode() {
        if (!this.isSelected()) {
            return null;
        }

        return this.findNode(this.selection.startKey);
    }

    findEndNode() {
        if (!this.isSelected()) {
            return null;
        }

        return this.findNode(this.selection.endKey);
    }

    prevNode(config) {
        if (!config) {
            return null;
        }

        let {col, row} = config;
        let node;
        if (col === 0) {
            row -= 1;
            const prevLine = this.lines[row];
            if (!prevLine) {
                node = null;
            }
            else {
                col = prevLine.length - 1;
                node = prevLine[col] || null;
            }
        }
        else {
            col -= 1;
            node = this.lines[row][col];
        }
        return node ? {node, row, col} : null;
    }

    nextNode(config) {
        if (!config) {
            return null;
        }

        let {col, row} = config;
        let node;
        const currentLine = this.lines[row];
        if (col === currentLine.length - 1) {
            row += 1;
            col = 0;
            const nextLine = this.lines[row];
            node = nextLine && nextLine[col] || null;
        }
        else {
            col += 1;
            node = currentLine[col];
        }
        return node ? {node, row, col} : null;
    }

    createText(text) {
        return {
            kind: 'text',
            key: this.generateKey(),
            text: text || ''
        };
    }

    lastLine() {
        return this.lines[this.lines.length - 1];
    }
    lastNode() {
        const lastLine = this.lastLine();
        return lastLine[lastLine.length - 1];
    }

    generateKey() {
        return (++keyCount) + '';
    }

    appendText(text) {
        const lastNode = this.lastNode();
        if (lastNode.kind === 'text') {
            lastNode.text += text;
            return lastNode;
        }

        const newTextNode = this.createText(text);
        this.lastLine().push(newTextNode);
        return newTextNode;
    }

    appendNode(node) {
        node = cloneDeep(node);
        if (this.lastNode().kind !== 'text') {
            this.appendText();
        }
        node.key = this.generateKey();
        this.lastLine().push(node);
    }

    appendNewLine() {
        this.lines.push([this.createText()]);
    }

    beforeInsert(noSplitText) {
        if (!this.isSelected()) {
            return;
        }

        if (!this.isCollapsed()) {
            // 删除选中
            this.deleteSelection();
        }

        const {startKey, startOffset} = this.selection;
        const start = this.findNode(startKey);
        if (!start) {
            // NOTE 不可能出现的情况
            return;
        }

        // 拆分文本
        if (!noSplitText && start.node.kind === 'text') {
            const oldText = start.node.text;
            start.node.text = oldText.slice(0, startOffset);
            const newTextNode = this.createText(oldText.slice(startOffset));
            this.lines[start.row].splice(start.col + 1, 0, newTextNode);
        }

        return start;
    }

    insertNewLine() {
        const start = this.beforeInsert();
        if (!start) {
            return;
        }

        // 拆分新行
        const prevLine = this.lines[start.row];
        const newLine = prevLine.splice(start.col + 1, prevLine.length - start.col - 1);
        if (!newLine[0] || newLine[0].kind !== 'text') {
            // 头部插入一个空字符串
            newLine.unshift(this.createText());
        }
        this.lines.splice(start.row + 1, 0, newLine);

        // 更新光标
        this.setCaret(newLine[0].key, 0);
    }

    insertText(text) {
        const start = this.beforeInsert(true);
        if (!start) {
            return;
        }

        const {startOffset} = this.selection;
        if (start.node.kind === 'text') {
            // 在文本中插入文本
            const startText = start.node.text;
            if (!startText) {
                this.replaceNodeKey(start.node, this.selection);
            }
            start.node.text = startText.slice(0, startOffset) + text + startText.slice(startOffset);
            // 更新光标
            this.setCaret(this.selection.startKey, this.selection.startOffset + text.length);
            return;
        }

        const line = this.lines[start.row];
        const nextNode = line[start.col + 1];
        if (nextNode.kind === 'text') {
            nextNode.text = text + nextNode.text;
            this.setCaret(nextNode.key, text.length);
        }
        else {
            const newTextNode = this.createText(text);
            line.splice(start.col + 1, 0, newTextNode);
            this.setCaret(newTextNode.key, text.length);
        }
    }

    insertNode(node) {
        node.key = node.key || this.generateKey();
        const start = this.beforeInsert();
        if (!start) {
            return;
        }

        const line = this.lines[start.row];
        let nextNode = line[start.col + 1];
        line.splice(start.col + 1, 0, node);
        if (nextNode.kind !== 'text') {
            // 插入一个文字节点，可以让浏览器放置光标
            nextNode = this.createText();
            line.splice(start.col + 2, 0, nextNode);
        }

        this.setCaret(nextNode.key, 0);
    }

    del() {
        if (!this.isSelected()) {
            return;
        }

        if (!this.isCollapsed()) {
            this.deleteSelection();
            return;
        }

        const {startKey, startOffset} = this.selection;
        const start = this.findNode(startKey);
        if (!start) {
            return;
        }

        const line = this.lines[start.row];
        if (start.node.kind === 'text') {
            const oldText = start.node.text;
            if (startOffset === oldText.length) {
                // 文字最后
                if (start.col === line.length - 1) {
                    // 最后一个，则合并下一行
                    if (start.row !== this.lines.length - 1) {
                        // 不是最后一行
                        const targetLine = this.lines[start.row];
                        const srcLine = this.lines[start.row + 1];
                        targetLine.push.apply(targetLine, srcLine);
                        this.lines.splice(start.row + 1, 1);
                        this.normalize(start.row);
                    }
                }
                else {
                    line.splice(start.col + 1, 1);
                    this.normalize(start.row);
                }
            }
            else {
                // 文字中间
                start.node.text = oldText.slice(0, startOffset) + oldText.slice(startOffset + 1);
            }
            // 强制刷新选中
            this.selectionApplyed = false;
        }
        else {
            // 不是文本，则用空文本替换节点
            const newTextNode = this.createText();
            line.splice(start.col, 1, newTextNode);
            this.setCaretAtFirst(newTextNode);
            this.normalize();
        }
    }

    backspace() {
        if (!this.isSelected()) {
            return;
        }

        if (!this.isCollapsed()) {
            this.deleteSelection();
            return;
        }

        const {startKey, startOffset} = this.selection;
        const start = this.findNode(startKey);
        if (!start) {
            return;
        }

        const line = this.lines[start.row];
        if (start.node.kind === 'text') {
            const oldText = start.node.text;
            if (startOffset === 0) {
                // 文字最前
                // const nextNode = line[start.col + 1];
                if (start.col === 0) {
                    // 第一个，则合并前一行
                    if (start.row !== 0) {
                        // 不是第一行
                        const targetLine = this.lines[start.row - 1];
                        const srcLine = this.lines[start.row];
                        targetLine.push.apply(targetLine, srcLine);
                        this.lines.splice(start.row, 1);
                        this.normalize(start.row - 1);
                    }
                    // 强制刷新选中
                    this.selectionApplyed = false;
                }
                else {
                    line.splice(start.col - 1, 1);
                    this.setCaret(start.node.key, 0);
                    this.normalize(start.row);
                }
            }
            else {
                // 文字中间
                start.node.text = oldText.slice(0, startOffset - 1) + oldText.slice(startOffset);
                this.setCaret(this.selection.startKey, startOffset - 1);
            }
        }
        else {
            // 不是文本，则用空文本替换节点
            const newTextNode = this.createText();
            line.splice(start.col, 1, newTextNode);
            this.setCaretAtFirst(newTextNode);
            this.normalize();
        }
    }

    toRawValue() {
        const lastLineCount = this.lines.length - 1;
        function pushText(result, text) {
            const lastItem = result[result.length - 1];
            if (typeof lastItem === 'string') {
                result[result.length - 1] += text;
            }
            else {
                result.push(text);
            }
        }

        return this.lines.reduce((result, line, index) => {
            result = line.reduce((result, node) => {
                if (node.kind !== 'text') {
                    node = omit(node, ['key']);
                    Object.keys(node).forEach((key) => {
                        node[key] = cloneDeep(node[key]);
                    });
                    result.push(node);
                    return result;
                }

                pushText(result, node.text);
                return result;
            }, result);

            if (lastLineCount !== index) {
                pushText(result, '\n');
            }
            return result;
        }, []);
    }

    /**
     * 判断两个raw value是否一致
     * @param {Array|string} oldValue 旧值
     * @param {Array|string} newValue 新值
     * @return {boolean} 是否一致
     */
    isSameRawValue(oldValue, newValue) {
        if (oldValue === newValue) {
            return true;
        }

        const oldType = typeof oldValue;
        if (oldType !== typeof newValue) {
            // 类型不一致
            return false;
        }

        // 类型一致，但是非 ===

        if (oldType === 'string') {
            // 字符串且非 ===，则不是一致
            return false;
        }

        if (!Array.isArray(oldValue)) {
            // 类型都是对象
            return isEqual(oldValue, newValue);
        }

        // 类型都是数组
        function notEmpty(value) {
            // 移除空字符串，undefined，null
            return !!value;
        }
        oldValue = oldValue.filter(notEmpty);
        newValue = newValue.filter(notEmpty);
        return isEqual(oldValue, newValue);
    }

    /**
     * 重置为格式化之后的值
     * @param {(Array|string)=} rawValue visual格式的输入值
     */
    resetFromRawValue(rawValue) {
        const state = this;
        if (state.isSameRawValue(rawValue, state.toRawValue())) {
            // 相同的值则不会更新value
            return;
        }

        this.reset();

        function textToState(text) {
            text.split(/\r?\n/)
                .forEach((val, i) => {
                    if (i !== 0) {
                        state.appendNewLine();
                    }
                    state.appendText(val);
                });
        }

        if (typeof rawValue === 'string') {
            textToState(rawValue);
            return;
        }

        if (!Array.isArray(rawValue)) {
            return;
        }

        rawValue.forEach((val) => {
            if (typeof val === 'string') {
                textToState(val);
            }
            else {
                // 不可编辑的行内元素
                state.appendNode(val);
            }
        });
        if (state.lastNode().kind !== 'text') {
            state.appendText();
        }
    }

    /**
     * 保存当前的选中状态，在blur时使用
     */
    blur() {
        this.savedSnapshotSelection = this.snapshotSelection();
    }

    /**
     * 重新恢复当前的选中状态，在focus时使用
     */
    focus() {
        if (!this.savedSnapshotSelection) {
            this.selectAtLast();
        }
        else {
            this.applySelection(this.savedSnapshotSelection);
            this.savedSnapshotSelection = null;
        }
    }

    /**
     * 判断当前文本是否是空的
     * @return {boolean} 是否是空的
     */
    isEmpty() {
        if (!this.lines || !this.lines.length) {
            return true;
        }

        if (this.lines.length > 1) {
            return false;
        }

        const firstLine = this.lines[0];
        const firstItem = firstLine[0];
        return !firstItem || (firstItem.kind === 'text' && firstItem.text === '');
    }
}

/**
 * 格式化
 * @param {Array} rawValue visual格式的输入值
 * @return {EditorState} 格式化之后的值
 */
EditorState.fromRawValue = function (rawValue) {
    const state = new EditorState();
    state.resetFromRawValue(rawValue);
    return state;
};

history(EditorState.prototype);
