import {insertAfter} from 'vhtml-ui/src/utils/dom';

const BLOCK_ELEMENTS = {
    ADDRESS: 1,
    BLOCKQUOTE: 1,
    CENTER: 1,
    DIR: 1,
    DIV: 1,
    DL: 1,
    FIELDSET: 1,
    FORM: 1,
    H1: 1,
    H2: 1,
    H3: 1,
    H4: 1,
    H5: 1,
    H6: 1,
    HR: 1,
    ISINDEX: 1,
    MENU: 1,
    NOFRAMES: 1,
    OL: 1,
    P: 1,
    PRE: 1,
    TABLE: 1,
    UL: 1
};

/**
 * get text
 * @param {Node} node node节点
 * @return {string} text
 */
export function getText(node) {
    if ('value' in node) {
        return node.value;
    }
    return node.textContent || node.innerText;
}

/**
 * 移动从指定节点之后的所有其他兄弟节点到新的父元素
 * @param {Element} newParent 新的父元素
 * @param {Node} startNode 开始节点(包含它)
 */
export function moveChildrenTo(newParent, startNode) {
    let nextNode;
    while (startNode) {
        nextNode = startNode.nextSibling;
        newParent.appendChild(startNode);
        startNode = nextNode;
    }
}

/**
 * 拆分textNode
 * @param {Element} node 文本节点
 * @param {number} offset 从第几个字符开始拆分
 * @return {Element} 拆分之后的后一个文本节点
 */
export function splitText(node, offset) {
    const doc = node.ownerDocument;
    if (offset === node.nodeValue.length) {
        // IE下做特殊处理
        var next = doc.createTextNode('');
        return insertAfter(node, next);
    }

    // ie8下splitText不会跟新childNodes, 我们手动触发他的更新
    // 但是我们不用支持IE8，这里留着做提示
    return node.splitText(offset);
}

/**
 * 是否是块级元素
 * @param {Node} node 节点
 * @return {boolean} 是否是块级元素
 */
export function isBlockElement(node) {
    return !!(node.nodeName && BLOCK_ELEMENTS[node.nodeName.toUpperCase()]);
}

/**
 * 计算节点的类型，判断是否不可编辑
 * @param {Node} node 节点
 * @return {boolean} 是否不可编辑
 */
export function isUnEditable(node) {
    return node.nodeType === 1 && !node.isContentEditable;
}

/**
 * 计算节点的类型，用数字表示
 * -1   不是视觉可见的节点
 * 0    是不可编辑缘故
 * 1    是非空文本节点
 * 2    是块级元素
 * @param {Node} node 节点
 * @return {number} 节点类型
 */
export function calcNodeType(node) {
    if (isUnEditable(node)) {
        return 0;
    }

    if (node.nodeType === 3 && node.nodeValue) {
        return 1;
    }

    if (isBlockElement(node)) {
        return 2;
    }

    return -1;
}

/**
 * 从指定节点开始拆分祖先元素，此元素及其之后的兄弟节点都会被放到新拆分出来的元素中去
 * 并且会一直向上拆分，直到拆分完第一个遇到的块级元素
 * 这个函数通常用来拆分一行
 * @param {Element} startNode 从此节点开始拆分父元素
 * @return {Element} 新的块级祖先元素
 */
export function splitLine(startNode) {
    let newParent;
    let node = startNode;
    let parent = node.parentNode;
    while (parent) {
        newParent = parent.cloneNode();
        insertAfter(parent, newParent);
        moveChildrenTo(newParent, node);

        if (isBlockElement(parent)) {
            break;
        }
        node = newParent;
        parent = newParent.parentNode;
    }
    return newParent;
}

/**
 * 合并指定节点内的两个相邻的文本子节点
 * @param {Node} node 元素节点
 */
export function normalize(node) {
    if (!node) {
        return;
    }
    if (node.nodeType === 3) {
        while (node.nextSibling && node.nextSibling.nodeType === 3) {
            node.nodeValue += node.nextSibling.nodeValue;
            node.parentNode.removeChild(node.nextSibling);
        }
    }
    else {
        normalize(node.firstChild);
    }
    normalize(node.nextSibling);
}

/**
 * 合并两个节点，前提是都是文本节点，或者拥有相同的nodeName
 * 会将后一个节点合并到前一个节点，所以后一个节点会消失
 * @param {Node} beforeNode 前一个节点
 * @param {Node} afterNode 后一个节点
 * @return {boolean} 合并成功与否
 */
export function mergeNodes(beforeNode, afterNode) {
    if (beforeNode.nodeType === 3 && afterNode.nodeType === 3) {
        // 合并文本节点
        beforeNode.nodeValue += afterNode.nodeValue;
    }
    else if (beforeNode.nodeName === afterNode.nodeName) {
        // 合并前后的相同元素
        moveChildrenTo(beforeNode, afterNode.firstChild);
    }
    else {
        return false;
    }

    afterNode.parentNode.removeChild(afterNode);
    normalize(beforeNode);
    return true;
}

/**
 * 合并两行
 * @param {Node} blockNode 行节点
 * @param {boolean} isForward 向前或者向后合并
 * @return {boolean} 合并成功与否
 */
export function mergeLine(blockNode, isForward) {
    const method = isForward ? 'nextSibling' : 'previousSibling';
    const otherBlockNode = blockNode[method];
    if (!isBlockElement(otherBlockNode)) {
        return false;
    }

    clearBR(isForward ? blockNode : otherBlockNode);
    return isForward
        ? mergeNodes(blockNode, otherBlockNode)
        : mergeNodes(otherBlockNode, blockNode);
}

/**
 * 判断节点内部是否是br标签结尾
 * @param {Element} node 节点
 * @return {boolean} 是否是br标签结尾
 */
export function isEndWithBR(node) {
    return node.lastElementChild && node.lastElementChild.nodeName === 'BR';
}

/**
 * 如果节点内最后有<br>标签，则删除
 * @param {Element} node 节点
 */
export function clearBR(node) {
    if (!isEndWithBR(node)) {
        return;
    }

    node.removeChild(node.lastElementChild);
}

/**
 * 如果节点内最后没有<br>标签，则新增一个<br>标签，撑起高度
 * @param {Element} node 节点
 */
export function fillupBR(node) {
    if (isEndWithBR(node)) {
        return;
    }

    node.appendChild(node.ownerDocument.createElement('br'));
}

/**
 * 检测节点node在父节点中的索引位置
 * @param {Node} node 需要检测的节点对象
 * @return {number} 该节点在父节点中的位置
 */
export function getNodeIndex(node) {
    let preNode = node;
    let i = 0;
    while ((preNode = preNode.previousSibling)) {
        i++;
    }
    return i;
}
