/**
 * 在一定范围内向前或向后查找节点，深度优先，会查子节点和自己，如果没有还会查父级的兄弟节点
 * @param {Node} startNode 被查找的起始节点
 * @param {boolean} isForward 向前还是向后查找
 * @param {Element} root 查找范围，不能超过这个元素
 * @param {function(Node)} whatType 计算type值的函数，如果不是-1则停止查找
 * @return {Object} 被找到的节点，及其type值
 *                  {
 *                      node: node,
 *                      type: -1
 *                  }
 */
function find(startNode, isForward, root, whatType) {
    const sonMethod = isForward ? 'firstChild' : 'lastChild';
    const siblingMethod = isForward ? 'nextSibling' : 'previousSibling';
    function nextNode(node) {
        return node[sonMethod] || node[siblingMethod];
    }
    function upNode(node) {
        const parent = (node === root || node.parentNode === root) ? null : node.parentNode;
        return parent ? (parent[siblingMethod] || upNode(parent)) : null;
    }

    let node = startNode;
    while (node) {
        let type = whatType(node);
        if (type >= 0) {
            // 如果类型正确则返回
            return {
                type,
                node
            };
        }

        // 类型不对，则继续查找下一个子节点或兄弟
        let next = nextNode(node);
        if (!next) {
            // 如果没有子节点了和兄弟节点了，则查找父节点的兄弟节点
            next = upNode(node);
        }
        node = next;
    }
}

/**
 * 在一定范围内向后查找节点，深度优先，会查子节点和自己，如果没有还会查父级的兄弟节点
 * @param {Node} startNode 被查找的起始节点
 * @param {Element} root 查找范围，不能超过这个元素
 * @param {function(Node)} whatType 计算type值的函数，如果不是-1则停止查找
 * @return {Object} 被找到的节点，及其type值
 *                  {
 *                      node: node,
 *                      type: -1
 *                  }
 */
export function findNext(startNode, root, whatType) {
    return find(startNode, true, root, whatType);
}

/**
 * 在一定范围内向前查找节点，深度优先，会查子节点和自己，如果没有还会查父级的兄弟节点
 * @param {Node} startNode 被查找的起始节点
 * @param {Element} root 查找范围，不能超过这个元素
 * @param {function(Node)} whatType 计算type值的函数，如果不是-1则停止查找
 * @return {Object} 被找到的节点，及其type值
 *                  {
 *                      node: node,
 *                      type: -1
 *                  }
 */
export function findPrev(startNode, root, whatType) {
    return find(startNode, false, root, whatType);
}

/**
 * 在一定范围内向前或向后查找子节点，深度优先
 * @param {Node} root\ 被查找的父节点
 * @param {boolean} isForward 向前还是向后查找
 * @param {function(Node)} whatType 计算type值的函数，如果不是-1则停止查找
 * @return {Object} 被找到的节点，及其type值
 *                  {
 *                      node: node,
 *                      type: -1
 *                  }
 */
export function findChildren(root, isForward, whatType) {
    const sonMethod = isForward ? 'firstChild' : 'lastChild';
    const siblingMethod = isForward ? 'nextSibling' : 'previousSibling';
    const ancestors = [];
    function nextNode(node) {
        const son = node[sonMethod];
        if (son && node !== root) {
            ancestors.push(node);
        }
        return son || node[siblingMethod];
    }
    function upNode() {
        const parentNode = ancestors.pop();
        return parentNode ? (parentNode[siblingMethod] || upNode()) : null;
    }

    let node = nextNode(root);
    while (node) {
        let type = whatType(node);
        if (type >= 0) {
            // 如果类型正确则返回
            return {
                type,
                node
            };
        }

        // 类型不对，则继续查找下一个子节点或兄弟
        node = nextNode(node);
        if (!node) {
            // 如果没有子节点了和兄弟节点了，就查找父节点的兄弟节点
            node = upNode();
        }
    }
}

/**
 * 获得第一个文本子节点
 * @param {Element} root 任意根节点
 * @param {string} method 方法
 * @param {function(Element):boolean=} match 是否匹配
 * @param {boolean=} notRoot 不是根节点，不对外开放的接口，内部循环使用
 * @return {Node} 找到的节点
 */
export function deepest(root, method, match, notRoot) {
    if (notRoot && (match ? match(root) : !root[method])) {
        // 如果有match函数，则通过即返回
        // 如果没有match，则找到最后一个再返回
        return root;
    }

    const next = root[method];
    if (!next) {
        return;
    }
    return deepest(next, method, match, true);
}

/**
 * 获得第一个文本子节点
 * @param {Element} node 任意节点
 * @return {Node} 文本节点
 */
export function firstTextNode(node) {
    if (node.nodeType === 3) {
        return node;
    }

    const result = findChildren(node, true, (node) => {
        return node.nodeType === 3 ? 0 : -1;
    });
    return result ? result.node : null;
}

/**
 * 找到DOM树顺序上的下一个或上一个节点，如果没有同级元素，则找父级节点的同级元素
 * @param {Node} node 节点
 * @param {boolean} isForward 向前或向后
 * @param {Element} root 查找范围，不能超过这个元素
 * @return {Node} 找到的节点
 */
export function navigate(node, isForward, root) {
    const method = isForward ? 'nextSibling' : 'previousSibling';
    let next = node[method];
    while (!next) {
        const parent = node.parentNode;
        if (!parent || parent === root) {
            return;
        }
        node = parent;
        next = node[method];
    }
    return next;
}

/**
 * 找到DOM树顺序上的下一个节点，如果没有同级元素，则找父级节点的同级元素
 * @param {Node} node 节点
 * @param {Element} root 查找范围，不能超过这个元素
 * @return {Node} 找到的节点
 */
export function next(node, root) {
    return navigate(node, true, root);
}

/**
 * 找到DOM树顺序上的上一个节点，如果没有同级元素，则找父级节点的同级元素
 * @param {Node} node 节点
 * @param {Element} root 查找范围，不能超过这个元素
 * @return {Node} 找到的节点
 */
export function prev(node, root) {
    return navigate(node, false, root);
}
