import {getNodeIndex, mergeNodes, mergeLine, isEndWithBR, isBlockElement} from './dom';

/**
 * 更新选中范围
 * @param {Range} range 选中范围
 * @param {Window=} context window上下文
 */
export function updateRange(range, context) {
    const selection = (context || window).getSelection();
    selection.removeAllRanges();
    selection.addRange(range);
}

/**
 * 取到当前第一个选中范围，没有则返回空
 * @return {Range} 选中范围
 */
export function getFirstRange() {
    const selection = window.getSelection();
    if (!selection.getRangeAt || !selection.rangeCount) {
        return;
    }

    return selection.getRangeAt(0);
}

/**
 * 设置光标的位置
 * @param {Node} node 节点
 * @param {number} offset 位置
 */
export function setCaret(node, offset) {
    const doc = node.ownerDocument;
    const range = doc.createRange();
    range.setStart(node, offset);
    range.collapse(true);
    updateRange(range, doc.defaultView);
}

/**
 * 取消当前选中范围
 * @param {Element} root editor root element
 */
export function deselect(root) {
    const selection = (root ? root.ownerDocument.defaultView : window).getSelection();
    if (!selection.getRangeAt || !selection.rangeCount) {
        return;
    }

    const range = selection.getRangeAt(0);
    if (!isRangeInsideElement(range, root)) {
        return;
    }

    selection.removeAllRanges();
}

/**
 * 获取在编辑器中的选中的range
 * @param {Element} root 编辑器根节点
 * @return {Range} range 选中范围
 */
export function getRangeInsideElement(root) {
    const selection = (root ? root.ownerDocument.defaultView : window).getSelection();
    if (!selection.getRangeAt || !selection.rangeCount) {
        return;
    }

    const range = selection.getRangeAt(0);
    return isRangeInsideElement(range, root) ? range : null;
}

/**
 * 判断选中的范围是否在指定的元素内部
 * @param {Range} range 选中范围
 * @param {Element} root 指定的元素
 * @return {boolean} 是否在内部
 */
export function isRangeInsideElement(range, root) {
    if (!range) {
        return false;
    }

    let node = range.commonAncestorContainer;
    if (node.nodeType === 3) {
        node = node.parentNode;
    }
    return root.contains(node);
}

/**
 * 设置选中
 * @param {Node} start 选中区域的开始节点
 * @param {number} startOffset 开始节点位移
 * @param {Node} end 选中区域的结束节点
 * @param {number} endOffset 结束节点位移
 */
export function select(start, startOffset, end, endOffset) {
    const doc = start.ownerDocument;
    const range = doc.createRange();
    range.setStart(start, startOffset);
    range.setEnd(end, endOffset);
    updateRange(range, doc.defaultView);
}

/**
 * 设置光标的位置在指定节点之前
 * @param {Node} node 节点
 */
export function setCaretBefore(node) {
    const parent = node.parentNode;
    if (!parent) {
        return;
    }

    const offset = getNodeIndex(node);
    setCaret(parent, offset);
}

/**
 * 设置光标的位置在指定节点之后
 * @param {Node} node 节点
 */
export function setCaretAfter(node) {
    const parent = node.parentNode;
    if (!parent) {
        return;
    }

    const offset = getNodeIndex(node);
    setCaret(parent, offset + 1);
}

/**
 * 获取不可编辑的区域
 * @param {Node} node 节点
 * @param {Node} root 编辑器的整体范围的DOM
 * @return {Element} 不可编辑的元素
 */
function getUnEditableArea(node, root) {
    if (node === root) {
        return;
    }

    while (node.nodeType !== 1 || node.isContentEditable) {
        node = node.parentNode;
        if (node === root) {
            return;
        }
    }
    return node;
}

/**
 * 计算两个节点合并之后，光标的正确位置
 * 主要目标是为了处理文本节点合并（normalize）之后的正确光标位置
 * NOTE: 只有如下情况，节点才会被合并：
 *      1. 两个都是文本节点
 *      2. 两个是相同nodeName的节点
 * @param {Node} beforeNode 前一个节点
 * @param {Node} afterNode 后一个节点
 * @return {Object} 光标位置
 */
function calcCaretBeforeMerge(beforeNode, afterNode) {
    if (beforeNode.nodeType === 3 && afterNode.nodeType === 3) {
        // 两个文本节点合并
        return {
            container: beforeNode,
            offset: beforeNode.nodeValue.length
        };
    }

    // 块级元素之间的合并，会去掉前一个节点的最后的br标签，避免合并后（合并成一行）视觉上换行依旧保留
    const beforeLineIsEndWithBR = isBlockElement(beforeNode) && isEndWithBR(beforeNode);
    const lastNodeOfBeforeNode = beforeLineIsEndWithBR
        ? beforeNode.lastChild.previousSibling : beforeNode.lastChild;
    const firstNodeOfAfterNode = afterNode.firstChild;

    // 如果都存在且都是文本，则会被normalize之后合并
    const canBeNormalize = lastNodeOfBeforeNode && firstNodeOfAfterNode
        && lastNodeOfBeforeNode.nodeType === 3 && firstNodeOfAfterNode.nodeType === 3;
    const caretParent = canBeNormalize ? lastNodeOfBeforeNode : beforeNode;
    const caretOffset = canBeNormalize
        ? lastNodeOfBeforeNode.nodeValue.length
        : caretParent.childNodes.length - (beforeLineIsEndWithBR ? 1 : 0);

    return {
        container: caretParent,
        offset: caretOffset
    };
}

/**
 * 合并当前光标的前后两个节点
 * 如果节点是相同标签或者都是文本节点时才会合并
 * 如果都处于一个文本节点之上，则不合并
 * @param {Range} range 代表光标位置的range对象，必须是collapsed
 */
export function mergeBetweenCaret(range) {
    if (!range.collapsed || range.startContainer.nodeType === 3) {
        return;
    }

    const children = range.startContainer.childNodes;
    const beforeNode = children[range.startOffset - 1];
    const afterNode = children[range.startOffset];
    if (!beforeNode || !afterNode) {
        return;
    }

    const caret = calcCaretBeforeMerge(beforeNode, afterNode);
    const mergeSuccess = mergeNodes(beforeNode, afterNode);
    if (!mergeSuccess) {
        return;
    }

    // 更新光标
    setCaret(caret.container, caret.offset);
}

/**
 * 合并两行，并正确的更新光标
 * @param {Element} node 被合并的一行的DOM节点
 * @param {boolean} isForward 向前或向后合并
 */
export function mergeLineBetweenCaret(node, isForward) {
    const beforeLine = isForward ? node : node.previousSibling;
    const afterLine = isForward ? node.nextSibling : node;
    const caret = calcCaretBeforeMerge(beforeLine, afterLine);

    if (!mergeLine(node, isForward)) {
        return;
    }
    // 合并成功则更新光标
    setCaret(caret.container, caret.offset);
}

/**
 * 删除选中的，如果不可编辑区域被部分选中，则需要全部删除它
 * @param {Node} root 编辑器的整体范围的DOM
 */
export function deleteSelection(root) {
    let range = getFirstRange();
    if (!range) {
        return;
    }

    const startUnEditableNode = getUnEditableArea(range.startContainer, root);
    if (startUnEditableNode) {
        // 开始节点在不可编辑元素内，则扩展删除范围
        range.setStartBefore(startUnEditableNode);
    }
    const endUnEditableNode = getUnEditableArea(range.endContainer, root);
    if (endUnEditableNode) {
        range.setEndAfter(endUnEditableNode);
    }
    range.deleteContents();
    mergeBetweenCaret(range);
}
