/**
 * @file Validator
 * @author mmzhou
 */

/**
 * 获取数组和字符串的长度
 * @param {*} value 数组和字符串或其他
 * @param {boolean} trim 是否对字符串忽略前后空格和换行符
 * @return {number} 如果不是数组或字符串，则返回-1
 */
const getLength = function (value, trim) {
    const isString = typeof value === 'string';
    return isString
        // 不考虑前后的空格和换行符
        ? (trim ? value.trim().length : value.length)
        : Array.isArray(value) ? value.length : -1;
};
const DEFAULT_VALIDATES = {
    required(value, rule) {
        if (!rule.required) {
            return true;
        }

        const trim = rule.trim == null ? true : rule.trim;
        const length = getLength(value, trim);
        return !!(~length ? length : value);
    },
    max(value, rule) {
        return value <= rule.max;
    },
    min(value, rule) {
        return value >= rule.min;
    },
    between(value, rule) {
        const min = rule.between[0];
        const max = rule.between[1];
        return (value <= max) && (value >= min);
    },
    maxLength(value, rule) {
        const trim = rule.trim == null ? true : rule.trim;
        const length = getLength(value, trim);
        return ~length ? (length <= rule.maxLength) : false;
    },
    minLength(value, rule) {
        const trim = rule.trim == null ? true : rule.trim;
        const length = getLength(value, trim);
        return ~length ? (length >= rule.minLength) : false;
    },
    betweenLength(value, rule) {
        const maxLength = rule.betweenLength[1];
        const minLength = rule.betweenLength[0];
        const trim = rule.trim == null ? true : rule.trim;
        const length = getLength(value, trim);
        return ~length
            ? ((length <= maxLength) && (length >= minLength))
            : false;
    },
    equalTo(value, rule, allValue) {
        return value === allValue[rule.equalTo];
    },
    validate(value, rule, allValue) {
        if (typeof rule.validate === 'function') {
            return rule.validate(value, rule, allValue);
        }

        return rule.validate.test(value);
    },
    /*
    ajax(value, allValue, returnAsync, target) {
        let body = allValue;
        let queryType = typeof target.query;
        if (queryType === 'string') {
            body = target.query.split(',').reduce((obj, q) => {
                q = q.trim();
                obj[q] = allValue[q];
                return obj;
            }, {});
        }
        else if (queryType === 'function') {
            body = target.query(value, allValue);
        }

        request({
            url: target.url || target,
            method: target.method,
            body: body
        }).then(
            (response) => {
                const data = response.body;
                returnAsync(data.success ? null : new Error(data.message || data.msg));
            },
            (response) => {
                returnAsync();
            }
        );
    },
    */
    /*eslint-disable */
    // From https://html.spec.whatwg.org/multipage/forms.html#valid-e-mail-address
    // Retrieved 2014-01-14
    // If you have a problem with this implementation, report a bug against the above spec
    // Or use custom methods to implement your own email validation
    email: /^[a-zA-Z0-9.!#$%&'*+\/=?^_`{|}~-]+@[a-zA-Z0-9](?:[a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?(?:\.[a-zA-Z0-9](?:[a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?)*$/,
    // Copyright (c) 2010-2013 Diego Perini, MIT licensed
    // https://gist.github.com/dperini/729294
    // see also https://mathiasbynens.be/demo/url-regex
    // modified to allow protocol-relative URLs
    url: /^(?:(?:(?:https?|ftp):)?\/\/)?(?:\S+(?::\S*)?@)?(?:(?!(?:10|127)(?:\.\d{1,3}){3})(?!(?:169\.254|192\.168)(?:\.\d{1,3}){2})(?!172\.(?:1[6-9]|2\d|3[0-1])(?:\.\d{1,3}){2})(?:[1-9]\d?|1\d\d|2[01]\d|22[0-3])(?:\.(?:1?\d{1,2}|2[0-4]\d|25[0-5])){2}(?:\.(?:[1-9]\d?|1\d\d|2[0-4]\d|25[0-4]))|(?:(?:[a-z\u00a1-\uffff0-9]-*)*[a-z\u00a1-\uffff0-9]+)(?:\.(?:[a-z\u00a1-\uffff0-9]-*)*[a-z\u00a1-\uffff0-9]+)*(?:\.(?:[a-z\u00a1-\uffff]{2,})).?)(?::\d{2,5})?(?:[/?#]\S*)?$/i,
    ip: /^(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$/
    /*eslint-enable */
};
const RESERVED_RULE_NAME = 'message,trigger,submit,async';

/**
 * 验证单个规则
 * @param {Object} rule 多个或单个规则
 * @param {*} value 值
 * @param {Object} allValue 其他关联值
 * @return {Promise} promise
 */
function validateRule(rule, value, allValue) {
    for (let ruleName in rule) {
        if (!rule.hasOwnProperty(ruleName)
            || !DEFAULT_VALIDATES.hasOwnProperty(ruleName)      // 不是内置规则
            || (~RESERVED_RULE_NAME.indexOf(ruleName))) {       // 是特殊属性，不能用作规则名字
            continue;
        }

        const defaultRule = DEFAULT_VALIDATES[ruleName];
        const result = typeof defaultRule === 'function'
            ? defaultRule(value, rule, allValue)
            : defaultRule.test(value);

        if (!!result
            && (typeof result === 'object' || typeof result === 'function')
            && typeof result.then === 'function') {
            // is promise
            return result;
        }
        else if (result === false) {
            // is false
            return Promise.reject(new Error(rule.message));
        }

        // success
        return Promise.resolve();
    }
}

export default {
    utils: DEFAULT_VALIDATES,

    /**
     * 添加默认验证规则
     * @param {string} ruleName 验证规则名
     * @param {function(*, Object, Object)|RegExp} rule rule
     */
    addRule(ruleName, rule) {
        if (~RESERVED_RULE_NAME.indexOf(ruleName)) {
            throw new Error('Can\'t use "message, trigger, submit, async, custom" as rule name.');
        }
        DEFAULT_VALIDATES[ruleName] = rule;
    },

    /**
     * validate one or more rules
     * @param {Array.<Object>} rules one or more rules.
     *              [
     *                  {required: true, message: 'Email required'},
     *                  {email: true, message: 'Please input email format string'},
     *                  {regexp: /regexp/, message: 'Wrong format'}
     *              ]
     * @param {*} value value
     * @param {Object} allValue all values
     * @return {Promise} promise result
     */
    validate(rules, value, allValue) {
        return rules.reduce((promise, rule) => {
            return promise.then(function () {
                try {
                    // success and move to next rule.
                    return validateRule(rule, value, allValue);
                }
                catch (e) {
                    // throw programmic error to console
                    setTimeout(() => {
                        throw e;
                    }, 0);
                    return Promise.reject(e);
                }
            });
        }, Promise.resolve());
    }
};
