import * as domUtil from 'vhtml-ui/src/utils/dom';
import raf from 'vhtml-ui/src/utils/raf';
import * as schedule from './schedule';

const DEFAULT_PRIORITY = ['left', 'right', 'top', 'bottom'];
const REVERSE_DIRECTION = {
    left: 'right',
    right: 'left',
    top: 'bottom',
    bottom: 'top'
};
const TUNNING_ORDER = ['bottom', 'right', 'top', 'left'];

/**
 * Popup class
 * manage the position of popup
 * @class
 */
export default class Popup {
    /**
     * Popup constructor
     * @param {Eleemnt} popup popup element.
     * @param {Object=} options options.
     * @param {string|Element} options.arrow arrow element or selector.
     * @param {string} options.placement in format or `direction-align`
     *                          @eg: 'left', 'left-start', 'left-end',
     *                               'right', 'right-start', 'right-end',
     *                               'top', 'top-start', 'top-end',
     *                               'bottom', 'bottom-start', bottom-end'
     * @param {boolean} options.fixed is placement fixed or not,
     *                          if not fixed, then automatically adjust popup,
     *                          to make sure popup in boundary element.
     * @param {string|Element} options.boundary boundary element.
     *                          @eg: 'offsetParent', 'viewport', boundaryElement
     * @param {number} options.boundaryGap boundaryGap of popup to boundary element.
     * @param {Array.<string>} options.priority priority of placement when adjust popup.
     * @param {number} options.offset offset.
     * @param {boolean} options.eventsEnabled Whether events (resize, scroll) are initially enabled.
     */
    constructor(popup, options) {
        this.popup = popup;
        const arrow = (options && options.arrow) || '[x-arrow]';
        this.arrow = typeof arrow === 'string' ? popup.querySelector(arrow) : arrow;
        this.placement = (options && options.placement) || 'bottom';
        this.fixed = (options && options.fixed) || false;
        this.priority = DEFAULT_PRIORITY;
        this.setPriority(options && options.priority);
        this.eventsEnabled = (options && options.eventsEnabled) || true;
        this.offset = options && parseInt(options.offset, 10) || 0;
        // window, viewport, offsetParent
        this.boundaryTarget = options && options.boundary || 'viewport';
        this.boundaryGap = options && options.boundaryGap || 0;
        this.syncWidth = options && options.syncWidth || false;
        this.overlayKey = options && options.overlayKey;
        this.updateCallback = options && options.updateCallback;
        this.currentVisible = options && options.currentVisible;

        this._updating = false;
        this._updateDebounce = () => {
            if (this._updating) {
                return;
            }
            this._updating = true;
            raf(() => {
                this.update();
                this._updating = false;
            });
        };
    }

    setPriority(priority) {
        if (!priority) {
            return;
        }
        this.priority = typeof priority === 'string' ? priority.split(',') : priority;
    }

    /**
     * bind trigger
     * @param {Element} trigger trigger element.
     */
    bind(trigger) {
        if (this.trigger && this.trigger !== trigger) {
            this._unbindTriggerEvents();
        }
        this.trigger = trigger;
        this.scrollParent = domUtil.getScrollParent(trigger);
        this._bindTriggerEvents();
    }

    /**
     * unbind trigger
     */
    unbind() {
        if (this.trigger) {
            this._unbindTriggerEvents();
        }
        this.trigger = null;
        this.scrollParent = null;
        this._unbindEvents();
    }

    /**
     * destroy
     */
    destroy() {
        if (typeof this.updateCallback === 'function') {
            this.updateCallback(this.overlayKey);
        }
        this.unbind();
        if (this.popup.parentNode) {
            this.popup.parentNode.removeChild(this.popup);
        }
    }

    /**
     * update position
     */
    update() {
        if (!this.trigger) {
            return;
        }
        // popper隐藏时，不需要重新计算位置
        if (!this.currentVisible) {
            if (typeof this.updateCallback === 'function') {
                this.updateCallback(this.overlayKey);
            }
            return;
        }

        this._update();

        if (this.eventsEnabled) {
            this._bindEvents();
        }
    }

    _update() {
        const triggerRect = domUtil.getRect(this.trigger);
        if (this.syncWidth) {
            this.popup.style.width = `${triggerRect.width}px`;
        }

        let popupRect = this._calcRect(this.placement, triggerRect);
        if (!this.fixed) {
            const adjustedPopupRect = this._adjustRect(triggerRect, popupRect);
            if (adjustedPopupRect) {
                // use adjusted popupRect instead
                popupRect = adjustedPopupRect;
            }
            else {
                // restore [x-placement]
                this.popup.setAttribute('x-placement', this.placement);
            }
        }
        // 该方法用于统一管理浮层
        if (typeof this.updateCallback === 'function') {
            this.updateCallback(this.overlayKey, popupRect);
        }
        this._apply(popupRect);
    }

    // apply style to popup
    _apply(popupRect) {
        const offsetRect = domUtil.getRect(domUtil.getOffsetParent(this.popup));
        this.popup.style.top = (popupRect.top - offsetRect.top) + 'px';
        this.popup.style.left = (popupRect.left - offsetRect.left) + 'px';

        if (this.arrow) {
            const arrowCSS = popupRect.arrowCSS;
            this.arrow.style[arrowCSS] = popupRect.arrow + 'px';
            this.arrow.style[arrowCSS === 'top' ? 'left' : 'top'] = '';
        }
    }

    _getBoundaryRect() {
        let boundary;
        switch (this.boundaryTarget) {
            case 'offsetParent':
                boundary = domUtil.getOffsetParent(this.popup);
                break;
            case 'viewport':
                boundary = 'viewport';
                break;
            case 'html':
                boundary = 'html';
                break;
            default:
                // specified boundary element
                boundary = this.boundaryTarget;
                break;
        }

        return domUtil.getRect(boundary);
    }

    _parsePlacement(placement) {
        const parts = placement.split('-');
        parts[1] = parts[1] || 'center';
        return parts;
    }

    _adjustRect(triggerRect, popupRect) {
        const boundaryRect = this._getBoundaryRect();
        const outsideDirection = domUtil.isOutside(popupRect, boundaryRect, this.boundaryGap);
        if (!outsideDirection) {
            // no need to adjust content
            return popupRect;
        }

        const [defaultDirection, defaultAlign] = this._parsePlacement(this.placement);
        // try other placements
        let result;
        this.priority.some((placement) => {
            if (placement === 'tunning') {
                result = this._tunningRect(triggerRect, popupRect, boundaryRect);
                return true;
            }

            const [thisDirection, thisAlign] = this._parsePlacement(placement);
            if (defaultDirection === thisDirection && defaultAlign === thisAlign) {
                // same as the default placement, then ignore
                return false;
            }

            const newPopupRect = this._calcRect(`${thisDirection}-${thisAlign}`, triggerRect);
            if (domUtil.isOutside(newPopupRect, boundaryRect, this.boundaryGap)) {
                // not inside boundary
                return false;
            }

            result = newPopupRect;
            return true;
        });
        return result;
    }

    _tunningRect(triggerRect, popupRect, boundaryRect) {
        const gap = this.boundaryGap;
        TUNNING_ORDER.forEach(function (direction) {
            const isTopOrLeft = direction === 'top' || direction === 'left';
            const reverseDirection = REVERSE_DIRECTION[direction];
            const boundary = Math[isTopOrLeft ? 'min' : 'max'](
                isTopOrLeft ? boundaryRect[direction] + gap : boundaryRect[direction] - gap,
                triggerRect[reverseDirection]
            );
            const offset = boundary - popupRect[direction];
            const needUpdate = isTopOrLeft ? offset > 0 : offset < 0;
            if (needUpdate) {
                popupRect[direction] = boundary;
                popupRect[reverseDirection] += offset;
            }
        });
    }

    _calcRect(placement, triggerRect) {
        // make sure the css style is right, in case of:
        // [x-placement^=left] { margin-right: 10px }
        this.popup.setAttribute('x-placement', placement);
        const popupRect = domUtil.getWidthAndHeight(this.popup);
        if (popupRect.width === 0 || popupRect.height === 0) {
            placement = '';
        }

        const parts = placement.split('-');
        const align = parts[1] || 'center';
        switch (placement) {
            case 'top':
            case 'top-center':
            case 'top-start':
            case 'top-end':
                popupRect.top = triggerRect.top - popupRect.height;
                popupRect.bottom = popupRect.top + popupRect.height;
                this._calcLeft(align, triggerRect, popupRect);
                break;
            case 'bottom':
            case 'bottom-center':
            case 'bottom-start':
            case 'bottom-end':
                popupRect.top = triggerRect.top + triggerRect.height;
                popupRect.bottom = popupRect.top + popupRect.height;
                this._calcLeft(align, triggerRect, popupRect);
                break;
            case 'left':
            case 'left-center':
            case 'left-start':
            case 'left-end':
                popupRect.left = triggerRect.left - popupRect.width;
                popupRect.right = popupRect.left + popupRect.width;
                this._calcTop(align, triggerRect, popupRect);
                break;
            case 'right':
            case 'right-center':
            case 'right-start':
            case 'right-end':
                popupRect.left = triggerRect.left + triggerRect.width;
                popupRect.right = popupRect.left + popupRect.width;
                this._calcTop(align, triggerRect, popupRect);
                break;
            default:
                popupRect.top = 0;
                popupRect.bottom = popupRect.top + popupRect.height;
                popupRect.left = 0;
                popupRect.right = popupRect.left + popupRect.width;
                // TODO
                popupRect.arrowCSS = 'left';
                popupRect.arrow = 0;
                break;
        }
        return popupRect;
    }

    _calcLeft(align, triggerRect, popupRect) {
        switch (align) {
            case 'center':
                popupRect.left = triggerRect.left + (triggerRect.width - popupRect.width) / 2 + this.offset;
                break;
            case 'start':
                popupRect.left = triggerRect.left + this.offset;
                break;
            case 'end':
                popupRect.left = triggerRect.left + triggerRect.width - popupRect.width - this.offset;
                break;
            default:
                popupRect.left = 0;
                break;
        }
        popupRect.right = popupRect.left + popupRect.width;

        popupRect.arrowCSS = 'left';
        if (!this.arrow) {
            popupRect.arrow = 0;
            return;
        }

        // update arrow position
        if (popupRect.left > triggerRect.right || triggerRect.left > popupRect.right) {
            // no overlapping then place arrow at center of popup
            popupRect.arrow = popupRect.width / 2;
            return;
        }

        // has overlapping then place arrow at center of overlapping area
        const overlappingLeft = Math.max(popupRect.left, triggerRect.left);
        const overlappingRight = Math.min(popupRect.right, triggerRect.right);
        popupRect.arrow = overlappingLeft + (overlappingRight - overlappingLeft) / 2 - popupRect.left;
    }

    _calcTop(align, triggerRect, popupRect) {
        switch (align) {
            case 'center':
                popupRect.top = triggerRect.top + (triggerRect.height - popupRect.height) / 2 + this.offset;
                break;
            case 'start':
                popupRect.top = triggerRect.top + this.offset;
                break;
            case 'end':
                popupRect.top = triggerRect.top + triggerRect.height - popupRect.height - this.offset;
                break;
            default:
                popupRect.top = 0;
                break;
        }
        popupRect.bottom = popupRect.top + popupRect.height;

        popupRect.arrowCSS = 'top';
        if (!this.arrow) {
            popupRect.arrow = 0;
            return;
        }

        // update arrow position
        if (popupRect.top > triggerRect.bottom || triggerRect.top > popupRect.bottom) {
            // no overlapping then place arrow at center of popup
            popupRect.arrow = popupRect.height / 2;
            return;
        }

        // has overlapping then place arrow at center of overlapping area
        const overlappingTop = Math.max(popupRect.top, triggerRect.top);
        const overlappingBottom = Math.min(popupRect.bottom, triggerRect.bottom);
        popupRect.arrow = overlappingTop + (overlappingBottom - overlappingTop) / 2 - popupRect.top;
    }

    _bindEvents() {
        if (this._bindedEvents) {
            return;
        }
        this._bindedEvents = true;
        schedule.bindEvent(window, this._updateDebounce);
    }
    _unbindEvents() {
        this._bindedEvents = false;
        schedule.unbindEvent(window, this._updateDebounce);
    }
    _bindTriggerEvents() {
        if (!this.scrollParent) {
            return;
        }
        schedule.bindEvent(this.scrollParent, this._updateDebounce);
    }
    _unbindTriggerEvents() {
        if (!this.scrollParent) {
            return;
        }
        schedule.unbindEvent(this.scrollParent, this._updateDebounce);
    }
}
