/**
 * 创建类似 v-popper 的指令
 *
 * Usage:
 *  v-popper.manual="somePopper"
 *  v-popper.show="somePopper"
 *  v-popper.hover="somePopper"
 *  v-popper.hover.delay="somePopper"
 *  v-popper.hover.delay1000="somePopper"
 *  v-popper.click="somePopper"
 *  v-popper.focus="somePopper"
 *
 *  v-popper:somePopper.manual
 *  v-popper:somePopper.show
 *  v-popper:somePopper.hover
 *  v-popper:somePopper.click
 *  v-popper:somePopper.focus
 *  v-popper:somePopper.show="data"
 *  v-popper:somePopper.hover="data"
 *  v-popper:somePopper.click="data"
 *  v-popper:somePopper.focus="data"
 */
import util from './util';
import {dashcase} from 'vhtml-ui/src/utils/lang';

function getPopper(context, refName, componentName, callback, notRetry) {
    let popper = context.$refs[refName];
    if (Array.isArray(popper)) {
        // 如果有多个，则用第一个
        // 有多个相同`ref`或者在`v-for`上使用`ref`会导致有多个
        popper = popper[0];
    }

    while (popper && popper.$options.name !== 'VPopper') {
        // 其他基于VPopper扩展的组件，可以再找一层，找到真正的`v-popper`
        popper = popper.$refs.popper;
    }

    if (popper) {
        callback(popper);
        return;
    }

    if (!popper && notRetry) {
        // 重试之后依旧找不到popper
        throw new Error(`[VHTML-UI][${dashcase(componentName)}] ${componentName} "${refName}" not found.`);
    }
    context.$nextTick(() => {
        // 在Vue 2.5之后，如下情况会导致ref拿不到（visible初始为false，后来改为true）
        // <template v-if="visible">
        //     <v-tip ref="btnnew" clazz="pop-right" placement="right">tipyu</v-tip>
        //     <span v-tip:btnnew.hover class="btn-new">
        //         <v-button kind="primary" disabled>按钮</v-button>
        //     </span>
        // </template>
        getPopper(context, refName, componentName, callback, true);
    });
}

function getDelay(modifiers) {
    const keys = Object.keys(modifiers);
    let key;
    keys.some((k) => {
        const isDelay = k.indexOf('delay') === 0;
        if (isDelay) {
            key = k;
        }
        return isDelay;
    });

    return key
        // delay === delay250
        ? parseInt(key.slice(5), 10) || 250
        : 0;
}

function createBindDirective(componentName, defaultMode, notRetry) {
    return function (el, binding, vnode) {
        const context = vnode.context;
        // dynamic value > argument
        const arg = binding.arg;
        const value = binding.value;
        const refName = arg || value;
        const modifiers = binding.modifiers;

        if (!modifiers || !refName || typeof refName !== 'string') {
            return;
        }

        let modeType;
        if (modifiers.focus) {
            // focus
            modeType = 'focus';
        }
        else if (modifiers.click) {
            // click
            modeType = 'click';
        }
        else if (modifiers.show) {
            // always show
            modeType = 'show';
        }
        else if (modifiers.manual) {
            // 手动处理，由组件自己调用show或hide方法实现
            modeType = 'manual';
        }
        else {
            // 默认值
            modeType = defaultMode;
        }

        getPopper(context, refName, componentName, function (popper) {
            if (popper.isDestroyed) {
                return;
            }
            // value is data, only when `v-popper:somePopper.show="data"`
            const data = arg ? value : null;
            util.bind(el, popper, modeType, getDelay(modifiers), data);
        }, notRetry);
    };
}

export default function createPopperDirective(componentName, defaultMode) {
    defaultMode = defaultMode || 'hover';
    const bindDirective = createBindDirective(componentName, defaultMode);
    const updateDirective = createBindDirective(componentName, defaultMode, true);
    return {
        bind: bindDirective,

        update(el, binding, vnode) {
            if (binding.oldValue === binding.value) {
                return;
            }

            updateDirective(el, binding, vnode);
        },

        unbind(el) {
            util.unbind(el);
        }
    };
}
