/**
 * Util
 */

const POPPER_HOOK_PROPNAME = 'VHTML_POPPER_HOOK_PROPNAME';

const util = {
    /**
     * bind trigger event
     * @param {Element} triggerElement trigger element
     * @param {Component} popper popup element
     * @param {string} modeType focus,click,hover,show,manual
     * @param {number=} delay delay hide. default is 0.
     * @param {*=} data popper data.
     */
    bind(triggerElement, popper, modeType, delay = 0, data) {
        popper.hideTimeout = null;
        const stopHide = () => {
            if (delay && popper.hideTimeout) {
                window.clearTimeout(popper.hideTimeout);
                popper.hideTimeout = null;
            }
        };
        const toggleCallback = function (e) {
            stopHide();
            popper.setTrigger(triggerElement, data);
            popper.toggle();
        };
        const showCallback = function (e) {
            stopHide();

            popper.setTrigger(triggerElement, data);
            popper.show();
        };
        const popperShowCallback = function (e) {
            if (popper.triggerElement && triggerElement !== popper.triggerElement) {
                return;
            }
            stopHide();

            popper.setTrigger(triggerElement, data);
            popper.show();
        };
        let hideCallback = function () {
            popper.hide();
            popper.setTrigger(triggerElement, data);
        };
        if (delay) {
            const trueHideCallback = hideCallback;
            hideCallback = function () {
                if (popper.triggerElement && triggerElement !== popper.triggerElement) {
                    // 不是当前的trigger在触发展示
                    return;
                }

                stopHide();
                popper.hideTimeout = window.setTimeout(trueHideCallback, delay);
            };
        }
        const bodyCallback = function (e) {
            if (!popper.currentVisible) {
                return;
            }
            const target = e.target;
            const popperElm = popper.$refs.popper;
            const triggerElm = triggerElement;
            if (popperElm.contains(target)              // 在popper中
                || triggerElm.contains(target)          // 在trigger中
                || popper.insideSubOverlay(target)      // 在子popper中
            ) {
                return;
            }

            popper.hide();
        };

        const defineValue = {
            modeType,
            popper
        };
        switch (modeType) {
            case 'focus':
                triggerElement.addEventListener('focus', showCallback, false);
                triggerElement.addEventListener('blur', hideCallback, false);
                defineValue.showCallback = showCallback;
                defineValue.hideCallback = hideCallback;
                break;
            case 'click':
                triggerElement.addEventListener('click', toggleCallback, false);
                document.body.addEventListener('click', bodyCallback, false);
                defineValue.toggleCallback = toggleCallback;
                defineValue.bodyCallback = bodyCallback;
                break;
            case 'hover':
                triggerElement.addEventListener('mouseenter', showCallback, false);
                triggerElement.addEventListener('mouseleave', hideCallback, false);
                if (delay) {
                    const popperElement = popper.$refs.popper;
                    popperElement.addEventListener('mouseenter', popperShowCallback, false);
                    popperElement.addEventListener('mouseleave', hideCallback, false);
                }
                defineValue.showCallback = showCallback;
                defineValue.hideCallback = hideCallback;
                defineValue.popperShowCallback = popperShowCallback;
                break;
            case 'show':
                // show
                showCallback();
                break;
            case 'manual':
                popper.setTrigger(triggerElement, data);
                break;
            default:
                break;
        }
        defineValue.unbindCallback = stopHide;

        const oldOptions = triggerElement[POPPER_HOOK_PROPNAME];
        if (oldOptions) {
            if (oldOptions.popper.currentVisible) {
                // if the old popper visible
                // then show the new popper
                popper.setTrigger(triggerElement);
                popper.show();
            }
            this.unbind(triggerElement);
            triggerElement[POPPER_HOOK_PROPNAME] = defineValue;
        }
        else {
            Object.defineProperty(triggerElement, POPPER_HOOK_PROPNAME, {
                value: defineValue,
                enumerable: false,
                configurable: true,
                writable: true
            });
        }

        // unbind on popper destroy
        popper.$on('beforeDestroy', () => {
            this.unbind(triggerElement);
        });
    },

    /**
     * bind trigger event
     * @param {Element} triggerElement trigger element
     */
    unbind(triggerElement) {
        /**
         * @type {Object} options
         * @type {Object} options.showCallback showCallback
         * @type {Object} options.hideCallback hideCallback
         * @type {Object} options.toggleCallback toggleCallback
         * @type {Object} options.bodyCallback bodyCallback
         * @type {string} options.modeType focus,click,hover,show
         */
        const options = triggerElement[POPPER_HOOK_PROPNAME];
        if (!options) {
            return;
        }

        switch (options.modeType) {
            case 'focus':
                triggerElement.removeEventListener('focus', options.showCallback);
                triggerElement.removeEventListener('blur', options.hideCallback);
                break;
            case 'click':
                triggerElement.removeEventListener('click', options.toggleCallback);
                document.body.removeEventListener('click', options.bodyCallback);
                break;
            case 'hover':
                const popperElement = options.popper.$refs.popper;
                triggerElement.removeEventListener('mouseenter', options.showCallback);
                triggerElement.removeEventListener('mouseleave', options.hideCallback);
                if (popperElement && options.popperShowCallback) {
                    popperElement.removeEventListener('mouseenter', options.popperShowCallback);
                    popperElement.removeEventListener('mouseleave', options.hideCallback);
                }
                break;
            default:
                break;
        }

        options.unbindCallback();
        // hide popup
        options.popper.hide();
        options.popper = null;
        triggerElement[POPPER_HOOK_PROPNAME] = null;
    }
};

export default util;
