const CHINESE_REGEXP = /[^\x00-\xff]/g;     // 匹配双字节字符，中文也是双字节的字符
                                            // [\u4e00-\u9fa5]，这个正则代表仅匹配中文，这里留作备用吧
const CALCULATORS = {
    default(value) {
        return this.password ? value.length : String(value).trim().length;
    },
    chinese(value) {
        return this.password ? value.length : value.replace(CHINESE_REGEXP, '00').length;
    }
};

/**
 * 这是一个用于处理最大可输入长度的mixin，目前用于input, textarea和editor
 * 使用此mixin的组件需要提供`currentValue`和`password`（如果是密码输入框，则为true）实现：`this.inputLength`
 */
export default {
    props: {
        maxlength: {
            type: [Number, String],
            default: -1
        },
        limit: {
            type: [Number, String],
            default: -1
        },
        lengthCalculator: {
            type: [Function, String],
            default: 'default'
        }
    },
    computed: {
        calcMaxlength() {
            let maxlength = this.maxlength;
            if (this.showLimit) {
                maxlength = Math.pow(10, this.limit.length + 1) - 1;
            }
            return maxlength;
        },
        showLimit() {
            return this.limit > 0;
        },
        inputLength() {
            const calculator = typeof this.lengthCalculator === 'string'
                ? CALCULATORS[this.lengthCalculator]
                : this.lengthCalculator;
            return calculator.call(this, this.currentValue);
        }
    },
    methods: {
        /**
         * 验证字数，用于兼容`form`组件
         * 验证成功则promise成功回调，否则则是promise失败回调
         * @param {*} value 当前值
         * @param {Object} rule 规则
         * @param {Object} allValue 所有表单内的其他值
         * @return {Promise} promise
         */
        validate(value, rule, allValue) {
            if (!this.showLimit) {
                return Promise.resolve();
            }

            if (this.inputLength > this.limit) {
                const ruleMessage = rule.message;
                return Promise.reject(new Error(
                    typeof ruleMessage === 'string'
                        ? ruleMessage.replace(/{limit}/g, this.limit)
                        : `不得超过${this.limit}个字符`
                ));
            }
            return Promise.resolve();
        }
    }
};
