export default {
    props: {
        multiple: Boolean,
        name: String,
        maps: {
            type: Object,
            default() {
                return {
                    text: 'text',
                    value: 'value',
                    disabled: 'disabled'
                };
            }
        },
        index: {
            type: [Number, Array],
            default() {
                return this.multiple ? [] : -1;
            }
        },
        value: {
            type: [Number, String, Array],
            default() {
                return this.multiple ? [] : null;
            }
        },
        options: {
            type: Array,
            default: () => []
        },
        strictMode: {
            type: Boolean,
            default: false
        },
        singleUnselect: Boolean
    },
    data() {
        const toValues = (result, i) => {
            const val = this.getValueByIndex(i);
            if (val == null) {
                return;
            }

            result.push(val);
            return result;
        };
        const hasValue = (typeof this.value === 'number' && this.value === this.value)
            || (this.value && this.value.length > 0);
        const currentValue = hasValue
            ? this.filterValue(this.value)
            : (this.multiple
                ? this.index.reduce(toValues, [])
                : this.getValueByIndex(this.index));

        let initByIndex = false;
        if (!hasValue) {
            initByIndex = this.multiple ? this.index.length > 0 : this.index !== -1;
        }

        return {
            currentValue,
            initByIndex
        };
    },
    watch: {
        value(nextValue) {
            this.currentValue = this.filterValue(nextValue);
        },
        index(nextValue) {
            this.currentValue = this.multiple
                ? nextValue.map((i) => {
                    return this.options[i][this.maps.value];
                })
                : this.options[nextValue][this.maps.value];
        },
        options(nextValue) {
            this.currentValue = this.filterValue(this.value);
        },
        currentValue(nextValue) {
            // events
            this.$emit('input', nextValue);
            this.$emit('change', nextValue);
        }
    },
    methods: {
        /**
         * Remove useless currentValue when options updated.
         * @param {number|string|Array} value value
         * @return {number|string|Array} new value.
         */
        filterValue(value) {
            if (!this.strictMode) {
                return this.value;
            }

            if (this.multiple) {
                const newValue = value.filter((val) => {
                    return this.options.some((opt) => !opt[this.maps.disabled] && opt[this.maps.value] === val);
                });
                return newValue.length === value.length ? value : newValue;
            }

            return this.options.some((opt) => !opt[this.maps.disabled] && opt[this.maps.value] === value)
                ? value : null;
        },
        getSelectedOptions(value) {
            return this.options.filter((opt) => {
                const selected = this.multiple
                        ? value.some((val) => opt[this.maps.value] === val)
                        : opt[this.maps.value] === value;
                return selected;
            });
        },
        getValueByIndex(index) {
            return this.options[index] ? this.options[index][this.maps.value] : null;
        },
        checkItemSelected(val) {
            return this.multiple
            ? this.currentValue.indexOf(val) > -1
            : this.currentValue === val;
        },
        handleSelect(evt, i) {
            const clickedOption = this.options[i];
            if (!clickedOption || clickedOption[this.maps.disabled]) {
                return;
            }

            if (this.multiple) {
                let selectedIdx = -1;
                const isSelected = this.currentValue.some((val, i) => {
                    if (val === clickedOption[this.maps.value]) {
                        selectedIdx = i;
                        return true;
                    }
                    return false;
                });

                if (isSelected) {
                    this.currentValue.splice(selectedIdx, 1);
                }
                else {
                    this.currentValue.push(clickedOption[this.maps.value]);
                }
            }
            else {
                this.currentValue = this.currentValue === clickedOption[this.maps.value]
                    ? (this.singleUnselect ? null : clickedOption[this.maps.value])
                    : clickedOption[this.maps.value];
            }

            // emit select event
            const selectedOptions = this.getSelectedOptions(this.currentValue);
            this.$emit('select', {
                value: this.currentValue,
                option: this.multiple ? selectedOptions
                    : selectedOptions[0] ? selectedOptions[0] : null,
                current: clickedOption
            });
        },

        clearSelection() {
            if (this.multiple) {
                this.currentValue.splice(0, this.currentValue.length);
            }
            else {
                this.currentValue = null;
            }
        }
    },
    created() {
        // sync value back, in case user set initial value by index
        if (this.initByIndex) {
            this.$emit('input', this.currentValue);
        }
    }
};
