import moment from 'moment';
/**
 * 将秒数格式化为时间形式
 * @param  {string | number} value 输入数据
 * @return {string}       格式化后的时间字符串
 */
const toTime = function (value) {
    value = parseInt(value, 10);
    let result = '-';
    if (value >= 0) {
        let second = value % 60;
        let secondS = second < 10 ? '0' + second : second;
        let minute = parseInt(value / 60, 10);
        let hour = parseInt(minute / 60, 10);
        minute = minute % 60;
        let minuteS = minute < 10 ? '0' + (minute % 60) : minute % 60;
        let hourS = hour < 10 ? '0' + hour : hour;
        result = (hour ? hourS + ':' : '') + minuteS + ':' + secondS;
    }
    return result;
};

/**
 * 格式化日期
 * @param  {number | string} value     日期的秒数形式
 * @param  {string} formatter 期望格式化的形式
 * @return {string}           格式化后的日期字符串
 */
const toDateString = function (value, formatter = 'YYYY-MM-DD') {
    return (value !== null && +value >= 0) ? moment.unix(+value).format(formatter) : '-';
};

/**
 * 格式化为百分比形式
 * @param  {string | number} data  输入数据
 * @param  {number}         digit 小数点后保留位数
 * @return {string}               格式化后的百分比字符串
 */
const toPercent = function (data, digit) {
    if (data !== null && +data >= 0) {
        digit = digit || 2;
        return (+data * 100).toFixed(digit) + '%';
    }
    return '-';
};

/**
 * 将数字转换为千分撇形式
 * @param  {string | number} data 输入数据
 * @return {string}      格式化后的字符串
 */
const toComma = function (data) {
    if (data !== null && +data >= 0) {
        data = data.toString();
        let numberRE = /(\d+)(\d{3})/g;
        while (numberRE.test(data)) {
            data = data.replace(numberRE, '$1' + ',' + '$2');
        }
        return data;
    }
    return '-';
};

const toTimeSpan = function (value) {
    let date = new Date(value * 1000);
    let h = date.getHours();
    return h + '时－' + (h + 1) + '时';
};

const toUnit = function (input) {
    if (input === undefined || input === '') {
        input = '-';
    }
    else if (input % 1 === 0) {
        input = input.toString();
        let numberRE = /(\d+)(\d{3})/g;
        let units = [{name: '亿', value: 99999999}, {name: '万', value: 99999}];
        let numberUnit = '';

        units.some(unit => {
            if (+input > unit.value) {
                if (unit.name === '万') {
                    numberUnit = ((+input) / (unit.value - 90000 + 1)).toFixed(1) + '';
                }
                else {
                    numberUnit = ((+input) / (unit.value + 1)).toFixed(1) + '';
                }
                numberUnit = numberUnit.replace(/\.0$/, '');
                numberUnit += `<span class="x-formatted-number-unit">${unit.name}</span>`;
                return true;
            }
        });

        if (numberUnit) {
            return numberUnit;
        }
        while (numberRE.test(input)) {
            input = input.replace(numberRE, '$1' + ',' + '$2');
        }
    }
    return input;
};

export default {
    toTime: toTime,
    toDateString: toDateString,
    toPercent: toPercent,
    toComma: toComma,
    toTimeSpan: toTimeSpan,
    toUnit: toUnit
};
